/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class Curve25519Donna */

#ifndef _Included_Curve25519Donna
#define _Included_Curve25519Donna
#ifdef __cplusplus
extern "C" {
#endif
/*
 * Class:     Curve25519Donna
 * Method:    curve25519Donna
 * Signature: ([B[B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_Curve25519Donna_curve25519Donna
  (JNIEnv *, jobject, jbyteArray, jbyteArray);

/*
 * Class:     Curve25519Donna
 * Method:    makePrivate
 * Signature: ([B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_Curve25519Donna_makePrivate
  (JNIEnv *, jobject, jbyteArray);

/*
 * Class:     Curve25519Donna
 * Method:    getPublic
 * Signature: ([B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_Curve25519Donna_getPublic
  (JNIEnv *, jobject, jbyteArray);

/*
 * Class:     Curve25519Donna
 * Method:    makeSharedSecret
 * Signature: ([B[B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_Curve25519Donna_makeSharedSecret
  (JNIEnv *, jobject, jbyteArray, jbyteArray);

/*
 * Class:     Curve25519Donna
 * Method:    helowrld
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_Curve25519Donna_helowrld
  (JNIEnv *, jobject);

#ifdef __cplusplus
}
#endif
#endif
